# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from wandb.proto import wandb_internal_pb2 as wandb_dot_proto_dot_wandb__internal__pb2
from wandb.proto import wandb_server_pb2 as wandb_dot_proto_dot_wandb__server__pb2
from wandb.proto import wandb_telemetry_pb2 as wandb_dot_proto_dot_wandb__telemetry__pb2


class InternalServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RunUpdate = channel.unary_unary(
                '/wandb_internal.InternalService/RunUpdate',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunUpdateResult.FromString,
                )
        self.Attach = channel.unary_unary(
                '/wandb_internal.InternalService/Attach',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.AttachRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.AttachResponse.FromString,
                )
        self.TBSend = channel.unary_unary(
                '/wandb_internal.InternalService/TBSend',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.TBRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.TBResult.FromString,
                )
        self.RunStart = channel.unary_unary(
                '/wandb_internal.InternalService/RunStart',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunStartRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunStartResponse.FromString,
                )
        self.GetSummary = channel.unary_unary(
                '/wandb_internal.InternalService/GetSummary',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryResponse.FromString,
                )
        self.SampledHistory = channel.unary_unary(
                '/wandb_internal.InternalService/SampledHistory',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryResponse.FromString,
                )
        self.PollExit = channel.unary_unary(
                '/wandb_internal.InternalService/PollExit',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PollExitRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PollExitResponse.FromString,
                )
        self.Shutdown = channel.unary_unary(
                '/wandb_internal.InternalService/Shutdown',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ShutdownRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ShutdownResponse.FromString,
                )
        self.RunExit = channel.unary_unary(
                '/wandb_internal.InternalService/RunExit',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunExitRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunExitResult.FromString,
                )
        self.RunPreempting = channel.unary_unary(
                '/wandb_internal.InternalService/RunPreempting',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingResult.FromString,
                )
        self.Metric = channel.unary_unary(
                '/wandb_internal.InternalService/Metric',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.MetricRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.MetricResult.FromString,
                )
        self.PartialLog = channel.unary_unary(
                '/wandb_internal.InternalService/PartialLog',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryResponse.FromString,
                )
        self.Log = channel.unary_unary(
                '/wandb_internal.InternalService/Log',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.HistoryRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.HistoryResult.FromString,
                )
        self.Summary = channel.unary_unary(
                '/wandb_internal.InternalService/Summary',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.SummaryRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.SummaryResult.FromString,
                )
        self.Config = channel.unary_unary(
                '/wandb_internal.InternalService/Config',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ConfigRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ConfigResult.FromString,
                )
        self.Files = channel.unary_unary(
                '/wandb_internal.InternalService/Files',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.FilesRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.FilesResult.FromString,
                )
        self.Output = channel.unary_unary(
                '/wandb_internal.InternalService/Output',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.OutputRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.OutputResult.FromString,
                )
        self.Telemetry = channel.unary_unary(
                '/wandb_internal.InternalService/Telemetry',
                request_serializer=wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryResult.FromString,
                )
        self.Alert = channel.unary_unary(
                '/wandb_internal.InternalService/Alert',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.AlertRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.AlertResult.FromString,
                )
        self.Artifact = channel.unary_unary(
                '/wandb_internal.InternalService/Artifact',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactResult.FromString,
                )
        self.LinkArtifact = channel.unary_unary(
                '/wandb_internal.InternalService/LinkArtifact',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactResult.FromString,
                )
        self.ArtifactSend = channel.unary_unary(
                '/wandb_internal.InternalService/ArtifactSend',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendResponse.FromString,
                )
        self.ArtifactPoll = channel.unary_unary(
                '/wandb_internal.InternalService/ArtifactPoll',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollResponse.FromString,
                )
        self.CheckVersion = channel.unary_unary(
                '/wandb_internal.InternalService/CheckVersion',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionResponse.FromString,
                )
        self.Pause = channel.unary_unary(
                '/wandb_internal.InternalService/Pause',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PauseRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PauseResponse.FromString,
                )
        self.Resume = channel.unary_unary(
                '/wandb_internal.InternalService/Resume',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ResumeRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ResumeResponse.FromString,
                )
        self.Status = channel.unary_unary(
                '/wandb_internal.InternalService/Status',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.StatusRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.StatusResponse.FromString,
                )
        self.ServerShutdown = channel.unary_unary(
                '/wandb_internal.InternalService/ServerShutdown',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownResponse.FromString,
                )
        self.ServerStatus = channel.unary_unary(
                '/wandb_internal.InternalService/ServerStatus',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerStatusRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerStatusResponse.FromString,
                )
        self.ServerInformInit = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformInit',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitResponse.FromString,
                )
        self.ServerInformStart = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformStart',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartResponse.FromString,
                )
        self.ServerInformFinish = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformFinish',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishResponse.FromString,
                )
        self.ServerInformAttach = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformAttach',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachResponse.FromString,
                )
        self.ServerInformDetach = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformDetach',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachResponse.FromString,
                )
        self.ServerInformTeardown = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformTeardown',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownResponse.FromString,
                )


class InternalServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RunUpdate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Attach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TBSend(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunStart(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSummary(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SampledHistory(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PollExit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Shutdown(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunExit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunPreempting(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Metric(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PartialLog(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Log(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Summary(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Config(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Files(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Output(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Telemetry(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Alert(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Artifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LinkArtifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ArtifactSend(self, request, context):
        """rpc messages for async operations: Send, Poll, Cancel, Release
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ArtifactPoll(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CheckVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Pause(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Resume(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Status(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerShutdown(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformInit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformStart(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformFinish(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformAttach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformDetach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformTeardown(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_InternalServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RunUpdate': grpc.unary_unary_rpc_method_handler(
                    servicer.RunUpdate,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunUpdateResult.SerializeToString,
            ),
            'Attach': grpc.unary_unary_rpc_method_handler(
                    servicer.Attach,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.AttachRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.AttachResponse.SerializeToString,
            ),
            'TBSend': grpc.unary_unary_rpc_method_handler(
                    servicer.TBSend,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.TBRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.TBResult.SerializeToString,
            ),
            'RunStart': grpc.unary_unary_rpc_method_handler(
                    servicer.RunStart,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunStartRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunStartResponse.SerializeToString,
            ),
            'GetSummary': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSummary,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryResponse.SerializeToString,
            ),
            'SampledHistory': grpc.unary_unary_rpc_method_handler(
                    servicer.SampledHistory,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryResponse.SerializeToString,
            ),
            'PollExit': grpc.unary_unary_rpc_method_handler(
                    servicer.PollExit,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PollExitRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PollExitResponse.SerializeToString,
            ),
            'Shutdown': grpc.unary_unary_rpc_method_handler(
                    servicer.Shutdown,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ShutdownRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ShutdownResponse.SerializeToString,
            ),
            'RunExit': grpc.unary_unary_rpc_method_handler(
                    servicer.RunExit,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunExitRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunExitResult.SerializeToString,
            ),
            'RunPreempting': grpc.unary_unary_rpc_method_handler(
                    servicer.RunPreempting,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingResult.SerializeToString,
            ),
            'Metric': grpc.unary_unary_rpc_method_handler(
                    servicer.Metric,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.MetricRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.MetricResult.SerializeToString,
            ),
            'PartialLog': grpc.unary_unary_rpc_method_handler(
                    servicer.PartialLog,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryResponse.SerializeToString,
            ),
            'Log': grpc.unary_unary_rpc_method_handler(
                    servicer.Log,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.HistoryRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.HistoryResult.SerializeToString,
            ),
            'Summary': grpc.unary_unary_rpc_method_handler(
                    servicer.Summary,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.SummaryRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.SummaryResult.SerializeToString,
            ),
            'Config': grpc.unary_unary_rpc_method_handler(
                    servicer.Config,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ConfigRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ConfigResult.SerializeToString,
            ),
            'Files': grpc.unary_unary_rpc_method_handler(
                    servicer.Files,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.FilesRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.FilesResult.SerializeToString,
            ),
            'Output': grpc.unary_unary_rpc_method_handler(
                    servicer.Output,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.OutputRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.OutputResult.SerializeToString,
            ),
            'Telemetry': grpc.unary_unary_rpc_method_handler(
                    servicer.Telemetry,
                    request_deserializer=wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryResult.SerializeToString,
            ),
            'Alert': grpc.unary_unary_rpc_method_handler(
                    servicer.Alert,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.AlertRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.AlertResult.SerializeToString,
            ),
            'Artifact': grpc.unary_unary_rpc_method_handler(
                    servicer.Artifact,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactResult.SerializeToString,
            ),
            'LinkArtifact': grpc.unary_unary_rpc_method_handler(
                    servicer.LinkArtifact,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactResult.SerializeToString,
            ),
            'ArtifactSend': grpc.unary_unary_rpc_method_handler(
                    servicer.ArtifactSend,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendResponse.SerializeToString,
            ),
            'ArtifactPoll': grpc.unary_unary_rpc_method_handler(
                    servicer.ArtifactPoll,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollResponse.SerializeToString,
            ),
            'CheckVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.CheckVersion,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionResponse.SerializeToString,
            ),
            'Pause': grpc.unary_unary_rpc_method_handler(
                    servicer.Pause,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PauseRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PauseResponse.SerializeToString,
            ),
            'Resume': grpc.unary_unary_rpc_method_handler(
                    servicer.Resume,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ResumeRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ResumeResponse.SerializeToString,
            ),
            'Status': grpc.unary_unary_rpc_method_handler(
                    servicer.Status,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.StatusRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.StatusResponse.SerializeToString,
            ),
            'ServerShutdown': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerShutdown,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownResponse.SerializeToString,
            ),
            'ServerStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerStatus,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerStatusRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerStatusResponse.SerializeToString,
            ),
            'ServerInformInit': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformInit,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitResponse.SerializeToString,
            ),
            'ServerInformStart': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformStart,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartResponse.SerializeToString,
            ),
            'ServerInformFinish': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformFinish,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishResponse.SerializeToString,
            ),
            'ServerInformAttach': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformAttach,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachResponse.SerializeToString,
            ),
            'ServerInformDetach': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformDetach,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachResponse.SerializeToString,
            ),
            'ServerInformTeardown': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformTeardown,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'wandb_internal.InternalService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class InternalService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RunUpdate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/RunUpdate',
            wandb_dot_proto_dot_wandb__internal__pb2.RunRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.RunUpdateResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Attach(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Attach',
            wandb_dot_proto_dot_wandb__internal__pb2.AttachRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.AttachResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TBSend(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/TBSend',
            wandb_dot_proto_dot_wandb__internal__pb2.TBRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.TBResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RunStart(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/RunStart',
            wandb_dot_proto_dot_wandb__internal__pb2.RunStartRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.RunStartResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetSummary(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/GetSummary',
            wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SampledHistory(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/SampledHistory',
            wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PollExit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/PollExit',
            wandb_dot_proto_dot_wandb__internal__pb2.PollExitRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.PollExitResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Shutdown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Shutdown',
            wandb_dot_proto_dot_wandb__internal__pb2.ShutdownRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ShutdownResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RunExit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/RunExit',
            wandb_dot_proto_dot_wandb__internal__pb2.RunExitRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.RunExitResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RunPreempting(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/RunPreempting',
            wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Metric(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Metric',
            wandb_dot_proto_dot_wandb__internal__pb2.MetricRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.MetricResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PartialLog(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/PartialLog',
            wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Log(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Log',
            wandb_dot_proto_dot_wandb__internal__pb2.HistoryRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.HistoryResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Summary(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Summary',
            wandb_dot_proto_dot_wandb__internal__pb2.SummaryRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.SummaryResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Config(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Config',
            wandb_dot_proto_dot_wandb__internal__pb2.ConfigRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ConfigResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Files(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Files',
            wandb_dot_proto_dot_wandb__internal__pb2.FilesRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.FilesResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Output(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Output',
            wandb_dot_proto_dot_wandb__internal__pb2.OutputRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.OutputResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Telemetry(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Telemetry',
            wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Alert(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Alert',
            wandb_dot_proto_dot_wandb__internal__pb2.AlertRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.AlertResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Artifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Artifact',
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LinkArtifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/LinkArtifact',
            wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ArtifactSend(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ArtifactSend',
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ArtifactPoll(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ArtifactPoll',
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CheckVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/CheckVersion',
            wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Pause(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Pause',
            wandb_dot_proto_dot_wandb__internal__pb2.PauseRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.PauseResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Resume(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Resume',
            wandb_dot_proto_dot_wandb__internal__pb2.ResumeRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ResumeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Status(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Status',
            wandb_dot_proto_dot_wandb__internal__pb2.StatusRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.StatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerShutdown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerShutdown',
            wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerStatus',
            wandb_dot_proto_dot_wandb__server__pb2.ServerStatusRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformInit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformInit',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformStart(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformStart',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformFinish(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformFinish',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformAttach(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformAttach',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformDetach(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformDetach',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformTeardown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformTeardown',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)
